#include "internal.h"

#if WITH_MPEG == 0
static jfauderr init(void) { return jfauderr_unrecognised; }
static jfauderr uninit(void) { return jfauderr_unrecognised; }
static jfauderr testfile(jfaud_file *file) { return jfauderr_unrecognised; }
static jfauderr cleanup(jfaud_file *file, char *noclose) { return jfauderr_unrecognised; }
static jfauderr readpcmdata(jfaud_file *file, jfaud_buffer *data, int samples) { return jfauderr_unrecognised; }
static jfauderr readallpcmdata(jfaud_file *file, jfaud_buffer *data) { return jfauderr_unrecognised; }
static jfauderr rewindpcmdata(jfaud_file *file) { return jfauderr_unrecognised; }
static jfauderr getplaytime(jfaud_file *file, double *len) { return jfauderr_unrecognised; }
#else

#include "mad.h"

#include <stdlib.h>
#include <stdio.h>

static struct jfaud_init_params *params;

#if (DYNAMIC_MPEG == 1)

#ifdef _WIN32
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#define DLLNAME "libMAD.dll"
#else
#include <dlfcn.h>
#define DLLNAME "libmad.so"
#endif

static struct {
#ifdef _WIN32
	HMODULE dll;
#else
	void *dll;
#endif
	char const *mad_version, *mad_copyright;
	char const *(*mad_stream_errorstr)(struct mad_stream const *);
	void (*mad_stream_init)(struct mad_stream *);
	void (*mad_stream_finish)(struct mad_stream *);
	void (*mad_stream_buffer)(struct mad_stream *, unsigned char const *, unsigned long);
	void (*mad_frame_init)(struct mad_frame *);
	void (*mad_frame_finish)(struct mad_frame *);
	int  (*mad_frame_decode)(struct mad_frame *, struct mad_stream *);
	void (*mad_synth_init)(struct mad_synth *);
	void (*mad_synth_frame)(struct mad_synth *, struct mad_frame const *);
} libmad = {NULL, NULL,NULL,NULL, NULL,NULL,NULL, NULL,NULL,NULL, NULL,NULL};

#define mad_version libmad.mad_version
#define mad_copyright libmad.mad_copyright
#define mad_stream_errorstr libmad.mad_stream_errorstr
#define mad_stream_init libmad.mad_stream_init
#define mad_stream_finish libmad.mad_stream_finish
#define mad_stream_buffer libmad.mad_stream_buffer
#define mad_frame_init libmad.mad_frame_init
#define mad_frame_finish libmad.mad_frame_finish
#define mad_frame_decode libmad.mad_frame_decode
#define mad_synth_init libmad.mad_synth_init
#define mad_synth_frame libmad.mad_synth_frame


#define testifloaded() if (libmad.dll == NULL) return jfauderr_notinited
#else
#define testifloaded() { }
#endif


#define INBUFSZ 20160
struct mpeginfo {
	struct mad_stream stream;
	struct mad_frame frame;
	struct mad_synth synth;

	unsigned long audiostart;

	unsigned char inbuf[INBUFSZ+MAD_BUFFER_GUARD];
	unsigned long inlen;
	
	unsigned long samplerate;
	unsigned short channels;

	unsigned short sampsused;
	unsigned short sampslen;
	signed short outbuf[2*1152];
};

#ifdef _MSC_VER
#define inline _inline
#endif
#ifdef __WATCOMC__
#define inline __inline
#endif

static jfauderr init(void)
{
	int er=0;
	void *sym;

	params = jfaud_getparams();

#if (DYNAMIC_MPEG == 1)
#ifdef _WIN32
	libmad.dll = LoadLibrary(DLLNAME);
	
#define gpa(d,x) sym = GetProcAddress(d,x); { \
		if (!sym) { \
			params->call_logmsg(jsprintf("Failed getting symbol %s from " DLLNAME,x)); \
			er = 1; \
		} \
	}
#else
	libmad.dll = dlopen(DLLNAME, RTLD_GLOBAL|RTLD_NOW);
	dlerror();

#define gpa(d,x) dlsym(d,x); { \
		const char *s; \
		if (s = dlerror()) { \
			params->call_logmsg(jsprintf("Failed getting symbol %s from " DLLNAME,x)); \
			er = 1; \
		} \
	}
#endif

	if (!libmad.dll) return jfauderr_notinited;
	
	mad_version = (char *)gpa(libmad.dll, "mad_version");
	mad_copyright = (char *)gpa(libmad.dll, "mad_copyright");
	er = 0;	// version and copyright aren't important
	mad_stream_errorstr = (void*)gpa(libmad.dll, "mad_stream_errorstr");
	mad_stream_init = (void*)gpa(libmad.dll, "mad_stream_init");
	mad_stream_finish = (void*)gpa(libmad.dll, "mad_stream_finish");
	mad_stream_buffer = (void*)gpa(libmad.dll, "mad_stream_buffer");
	mad_frame_init = (void*)gpa(libmad.dll, "mad_frame_init");
	mad_frame_finish = (void*)gpa(libmad.dll, "mad_frame_finish");
	mad_frame_decode = (void*)gpa(libmad.dll, "mad_frame_decode");
	mad_synth_init = (void*)gpa(libmad.dll, "mad_synth_init");
	mad_synth_frame = (void*)gpa(libmad.dll, "mad_synth_frame");

	if (er) {
#ifdef _WIN32
		FreeLibrary(libmad.dll);
#else
		dlclose(libmad.dll);
#endif
		libmad.dll = NULL;
		params->call_logmsg("Failed fetching required symbols from " DLLNAME ".");
		return jfauderr_notinited;
	}

	if (mad_version && mad_copyright) {
		params->call_logmsg(jsprintf(DLLNAME ": %s %s",mad_version,mad_copyright));
	}
#endif	
	
	return jfauderr_ok;
}

static jfauderr uninit(void)
{
#if (DYNAMIC_MPEG == 1)
	if (libmad.dll) {
#ifdef _WIN32
		FreeLibrary(libmad.dll);
#else
		dlclose(libmad.dll);
#endif
		libmad.dll = NULL;
	}
#endif
	return jfauderr_ok;
}

static jfauderr testfile(jfaud_file *file, const char *sub)
{
	struct mpeginfo mf, *mpegf;
	int ret;
	unsigned char header[10];
	long taglen;

	testifloaded();

	memset(&mf, 0, sizeof(struct mpeginfo));
	
	params->call_seek(file->fh, 0, SEEK_SET);
	if (params->call_read(file->fh, header, 10) != 10) return jfauderr_genericerror;

	// check if there's a fricking id3v2 tag at the beginning
	if (header[0] == 'I' && header[1] == 'D' && header[2] == '3') {
		// seems to be, validate it and see how much crap we need to seek past
		taglen = header[6]&0x7f;
		taglen = (taglen<<7)|(header[7]&0x7f);
		taglen = (taglen<<7)|(header[8]&0x7f);
		taglen = (taglen<<7)|(header[9]&0x7f);

		mf.audiostart = taglen+10;

		params->call_seek(file->fh, taglen, SEEK_CUR);
		if (params->call_read(file->fh, header, 4) != 4) return jfauderr_genericerror;
	}

	// test the header info to see whether it's recognisable as an mpeg audio file
	if (header[0] != 0xff || (header[1]&0xe0) != 0xe0) return jfauderr_unrecognised;
	if ((header[1]&0x18) == 0x8) return jfauderr_unrecognised;
	if ((header[1]&0x06) == 0) return jfauderr_unrecognised;
	if ((header[2]&0x0c) == 0xc) return jfauderr_unrecognised;
	// one would guess that everything following is mpeg data

	params->call_seek(file->fh, mf.audiostart, SEEK_SET);

	mpegf = (struct mpeginfo *)malloc(sizeof(struct mpeginfo));
	if (!mpegf) return jfauderr_outofmemory;

	memcpy(mpegf, &mf, sizeof(struct mpeginfo));

	mad_stream_init(&mpegf->stream);
	mad_frame_init(&mpegf->frame);
	mad_synth_init(&mpegf->synth);	

	file->type = jfaudfile_mpeg;
	file->readerinfo = (void*)mpegf;

	return jfauderr_ok;
}

static jfauderr cleanup(jfaud_file *file, char *noclose)
{
	testifloaded();

	if (file->type != jfaudfile_mpeg) return jfauderr_unrecognised;
	
	if (file->readerinfo) {
		struct mpeginfo *mpegf;
		mpegf = (struct mpeginfo *)file->readerinfo;

		mad_synth_finish(&mpegf->synth);
		mad_frame_finish(&mpegf->frame);
		mad_stream_finish(&mpegf->stream);

		free(file->readerinfo);
	}
	file->readerinfo = NULL;
	*noclose = 0;

	return jfauderr_ok;
}

static inline signed short madscalesamp(mad_fixed_t sample)
{
	// round
	sample += (1L << (MAD_F_FRACBITS - 16));

	// clip
	if (sample >= MAD_F_ONE) sample = MAD_F_ONE - 1;
	else if (sample < -MAD_F_ONE) sample = -MAD_F_ONE;

	// quantize
	return sample >> (MAD_F_FRACBITS + 1 - 16);
}

static jfauderr readpcmdata(jfaud_file *file, jfaud_buffer *data, int samples)
{
	struct mpeginfo *mpegf;
	int ret=0, sampsleft, ns, i;
	signed short *wp;
	jfauderr err;
	
	testifloaded();

	if (file->type != jfaudfile_mpeg) return jfauderr_unrecognised;

	mpegf = (struct mpeginfo *)file->readerinfo;

	sampsleft = samples;
	data->numsamples = 0;
	data->data = NULL;
	while (sampsleft > 0) {
		if (mpegf->sampslen - mpegf->sampsused == 0) {
			unsigned long readamt = 0;

			if (mpegf->stream.next_frame) {
				i = (long)mpegf->stream.bufend - (long)mpegf->stream.next_frame;
				memmove(mpegf->inbuf, mpegf->stream.next_frame, i);
				mpegf->inlen = i;
				readamt = INBUFSZ - i;
			} else {
				readamt = INBUFSZ;
			}

			ret = params->call_read(file->fh, &mpegf->inbuf[mpegf->inlen], readamt);
			if (ret == 0) return jfauderr_ok;	// eof
			if (ret < 0) return jfauderr_genericerror;

			mpegf->inlen += ret;
			memset(&mpegf->inbuf[mpegf->inlen], 0, INBUFSZ+MAD_BUFFER_GUARD-mpegf->inlen);

			if (ret < readamt) {
				// close to eof, add guard space
				memset(&mpegf->inbuf[mpegf->inlen],0,MAD_BUFFER_GUARD);
				mpegf->inlen += MAD_BUFFER_GUARD;
			}

			mad_stream_buffer(&mpegf->stream, mpegf->inbuf, mpegf->inlen);
			mpegf->stream.error = 0;

			if (mad_frame_decode(&mpegf->frame, &mpegf->stream)) {
				if (MAD_RECOVERABLE(mpegf->stream.error)) {
					if (mpegf->stream.error != MAD_ERROR_LOSTSYNC ||
					    ret == readamt ||
					    memcmp(mpegf->stream.this_frame,"TAG",3)) {
						debuglog(params->call_logmsg,
								jsprintf("file_mpeg: recoverable frame-level error (%s)",
									mad_stream_errorstr(&mpegf->stream)));
					}
				} else if (mpegf->stream.error == MAD_ERROR_BUFLEN) {
					// this shouldn't happen since we top up things every time
					debuglog(params->call_logmsg,"file_mpeg: need data!?!?");
					return jfauderr_genericerror;
				} else {
					debuglog(params->call_logmsg,
							jsprintf("file_mpeg: unrecoverable frame-level error (%s)",
							mad_stream_errorstr(&mpegf->stream)));
					return jfauderr_genericerror;
				}
			}

			mpegf->samplerate = mpegf->frame.header.samplerate;
			mpegf->channels = (mpegf->frame.header.mode > MAD_MODE_SINGLE_CHANNEL) + 1;

			mad_synth_frame(&mpegf->synth, &mpegf->frame);

			mpegf->sampslen = mpegf->synth.pcm.length;
			mpegf->sampsused = 0;
			if (mpegf->channels == 1) {
				for (i=mpegf->sampslen-1; i>=0; i--) {
					mpegf->outbuf[i] = madscalesamp(mpegf->synth.pcm.samples[0][i]);
				}
			} else {
				for (i=mpegf->sampslen-1; i>=0; i--) {
					mpegf->outbuf[(i<<1)] = madscalesamp(mpegf->synth.pcm.samples[0][i]);
					mpegf->outbuf[(i<<1)+1] = madscalesamp(mpegf->synth.pcm.samples[1][i]);
				}
			}
		}

		if (!data->data) {
			if ((err = jfaudbuffer_alloc(data, mpegf->samplerate, mpegf->channels, 2*8, samples)) != jfauderr_ok) return err;
			data->numsamples = 0;
		}
		wp = (signed short *)data->data + (data->numsamples*data->channels);

		ns = min(mpegf->sampslen - mpegf->sampsused, sampsleft);

		if (data->channels == 1) {
			for (i=ns-1; i>=0; i--) {
				*(wp++) = mpegf->outbuf[mpegf->sampsused];
				mpegf->sampsused++;
			}
		} else {
			for (i=ns-1; i>=0; i--) {
				*(wp++) = mpegf->outbuf[(mpegf->sampsused<<1)];
				*(wp++) = mpegf->outbuf[(mpegf->sampsused<<1)+1];
				mpegf->sampsused++;
			}
		}

		sampsleft -= ns;
		data->numsamples += ns;
	}
	
	return jfauderr_ok;
}

static jfauderr readallpcmdata(jfaud_file *file, jfaud_buffer *data)
{
	testifloaded();
	return jfauderr_genericerror;
}

static jfauderr rewindpcmdata(jfaud_file *file)
{
	struct mpeginfo *mpegf;
	
	testifloaded();
	
	mpegf = (struct mpeginfo *)file->readerinfo;

	params->call_seek(file->fh, mpegf->audiostart, SEEK_SET);
	mpegf->inlen = mpegf->sampslen = mpegf->sampsused = 0;
	
	// reset the decoder
	mad_synth_finish(&mpegf->synth);
	mad_frame_finish(&mpegf->frame);
	mad_stream_finish(&mpegf->stream);

	mad_stream_init(&mpegf->stream);
	mad_frame_init(&mpegf->frame);
	mad_synth_init(&mpegf->synth);	
	
	return jfauderr_ok;
}

static jfauderr getpcmlength(jfaud_file *file, unsigned int *samples)
{
	testifloaded();
	*samples = 3600 * 44100;	// Getting sample count is not quite simple
	return jfauderr_ok;
}

static jfauderr getplaytime(jfaud_file *file, double *len)
{
	testifloaded();
	*len = 3600.0;	// Getting the length of an MP3 is not quite a straightforward task
	return jfauderr_ok;
}

#endif

jfaud_filereader reader_mpeg = {
	"MPEG Audio",
	jfaudfile_mpeg,
	jfaudfiledata_pcm,

	init,
	uninit,
	testfile,
	cleanup,

	readpcmdata,
	readallpcmdata,
	rewindpcmdata,
	getpcmlength,
	getplaytime
};
