#include "internal.h"

#include <stdlib.h>
#include <stdio.h>

static struct jfaud_init_params *params;

#ifdef __GNUC__
#define PACKED __attribute__ ((packed))
#else
#define PACKED
#endif

struct vocinfo {
	unsigned long audiostart;
	unsigned long audiolength;
	unsigned long pcmposn;

	long blocksize;
	char blocktype;
	union {
#ifdef _MSC_VER
#pragma pack(1)
#endif
		struct PACKED {
			unsigned char tc;
			unsigned char compr;
		} block1;
		struct PACKED {
			unsigned short length;
			unsigned char tc;
		} block3;
		struct PACKED {
			unsigned short id;
		} block4;
		struct PACKED {
			unsigned short repeat;
		} block6;
		struct PACKED {
			unsigned short tc;
			unsigned char pack;
			unsigned char mode;
		} block8;
		struct PACKED {
			unsigned long samplerate;
			unsigned char bitspersample;
			unsigned char channels;
			unsigned short format;
			char filler[4];
		} block9;
#ifdef _MSC_VER
#pragma pack()
#endif
	} blockinfo;

	long samplerate;
	short bitspersample;
	short channels;
};

static jfauderr rewindpcmdata(jfaud_file *file);

static jfauderr init(void)
{
	params = jfaud_getparams();

	return jfauderr_ok;
}

static jfauderr uninit(void)
{
	return jfauderr_ok;
}

static jfauderr testfile(jfaud_file *file, const char *sub)
{
	char sig[20];
	short offset;
	short version, version2;
	struct vocinfo *voc;
	jfauderr err;
	int br;

	params->call_seek(file->fh, 0, SEEK_SET);

	if (params->call_read(file->fh, sig, 20) != 20) return jfauderr_genericerror;
	if (memcmp(sig, "Creative Voice File\x1a", 20)) return jfauderr_unrecognised;
	
	br  = params->call_read(file->fh, &offset, 2);
	br += params->call_read(file->fh, &version, 2);
	br += params->call_read(file->fh, &version2, 2);
	if (br != 6) return jfauderr_genericerror;

	if (~version + 0x1234 != version2) return jfauderr_unrecognised;

	voc = (struct vocinfo *)malloc(sizeof(struct vocinfo));
	voc->audiostart = offset;
	voc->audiolength = 0;
	voc->pcmposn = 0;

	voc->blocksize = 0;
	voc->blocktype = -1;

	voc->samplerate = -1;
	voc->channels = -1;
	voc->bitspersample = -1;

	err = jfauderr_ok;

	while (err == jfauderr_ok) {
		if (params->call_read(file->fh, &voc->blocktype, 1) != 1) { err = jfauderr_genericerror; break; }
		if (voc->blocktype == 0) break;	// terminator

		voc->blocksize = 0;
		if (params->call_read(file->fh, &voc->blocksize, 3) != 3) { err = jfauderr_genericerror; break; }

		switch (voc->blocktype) {
			case 1:
				if (params->call_read(file->fh, &voc->blockinfo, 2) != 2)
					{ err = jfauderr_genericerror; break; }
				voc->blocksize -= 2;	// snip off the header
				if (voc->blockinfo.block1.compr != 0)
					{ err = jfauderr_genericerror; break; }
				if (voc->samplerate < 0) {
					voc->samplerate = 1000000/(256-voc->blockinfo.block1.tc); //tc=256-(1000000/sample_rate)
					voc->channels = 1;
					voc->bitspersample = 8;
				} else {
					// we don't like the sample rate to change
					if (voc->samplerate != 1000000/(256-voc->blockinfo.block1.tc))
						{ err = jfauderr_genericerror; break; }
				}
				voc->audiolength += voc->blocksize;
				break;
			case 2:
				// missing a preceeding block 1 or 8
				if (voc->samplerate < 0) { err = jfauderr_genericerror; break; }
				voc->audiolength += voc->blocksize;
				break;
			case 8:
				if (params->call_read(file->fh, &voc->blockinfo, 4) != 4)
					{ err = jfauderr_genericerror; break; }
				voc->blocksize -= 4;	// snip off the header
				if (voc->samplerate < 0) {
					voc->samplerate = 256000000/(65536-voc->blockinfo.block8.tc); //tc=65536-(256000000/sample_rate)
					voc->samplerate >>= voc->blockinfo.block8.mode;
					voc->channels = voc->blockinfo.block8.mode + 1;
					voc->bitspersample = 8;
				} else {
					// we don't like the sample format changing
					{ err = jfauderr_genericerror; break; }
				}
				break;
			case 9:
				if (params->call_read(file->fh, &voc->blockinfo, 12) != 12)
					{ err = jfauderr_genericerror; break; }
				voc->blocksize -= 12;	// snip off the header
				voc->samplerate = voc->blockinfo.block9.samplerate;
				voc->channels = voc->blockinfo.block9.channels;
				voc->bitspersample = voc->blockinfo.block9.bitspersample;
				voc->audiolength += voc->blocksize;
				break;
			default: break;
		}
		if (voc->blocksize > 0)
			params->call_seek(file->fh, voc->blocksize, SEEK_CUR);
	}
	
	if (err != jfauderr_ok) {
		free(voc);
		return err;
	}
	file->type = jfaudfile_voc;
	file->readerinfo = (void*)voc;
	rewindpcmdata(file);
	
	return jfauderr_ok;
}

static jfauderr cleanup(jfaud_file *file, char *noclose)
{
	if (file->readerinfo) free(file->readerinfo);
	file->readerinfo = NULL;
	*noclose = 0;

	return jfauderr_ok;
}

static jfauderr readvocsamples(jfaud_file *file, jfaud_buffer *data, int maxsamples)
{
	char *bp;
	int br,rbr;
	struct vocinfo *voc;
	int maxbytes;

	voc = (struct vocinfo *)file->readerinfo;
	bp = (char*)data->data;
	data->numsamples = 0;

	maxbytes = maxsamples * voc->channels * voc->bitspersample / 8;

	while (maxbytes > 0) {
		if (voc->blocktype == -1) {
			// get the next block
			if (params->call_read(file->fh, &voc->blocktype, 1) != 1) { return jfauderr_genericerror; }
			if (voc->blocktype != 0) {
				voc->blocksize = 0;
				if (params->call_read(file->fh, &voc->blocksize, 3) != 3) { return jfauderr_genericerror; }
			} else {
				data->numsamples = ((long)bp - (long)data->data) / (voc->channels * voc->bitspersample / 8);
				return jfauderr_ok;
			}
			switch (voc->blocktype) {
				case 1:
					if (params->call_read(file->fh, &voc->blockinfo, 2) != 2)
						{ return jfauderr_genericerror; }
					voc->blocksize -= 2;	// snip off the header
					break;
				case 2: break;
				case 9:
					if (params->call_read(file->fh, &voc->blockinfo, 12) != 12)
						{ return jfauderr_genericerror; }
					voc->blocksize -= 12;	// snip off the header
					break;
				default: break;
			}
				
		}
		switch (voc->blocktype) {
			case 1:
			case 2:
			case 9:
				br = min(voc->blocksize, maxbytes);
				rbr = params->call_read(file->fh, bp, br);
				bp += rbr;
				voc->pcmposn += rbr;
				voc->blocksize -= rbr;
				maxbytes -= rbr;
				if (rbr < br) return jfauderr_genericerror;
				break;
			default:
				// skip the unused block
				params->call_seek(file->fh, voc->blocksize, SEEK_CUR);
				voc->blocksize = 0;
				break;
		}
		if (voc->blocksize == 0) voc->blocktype = -1;
	}

	data->numsamples = ((long)bp - (long)data->data) / (voc->channels * voc->bitspersample / 8);
	return jfauderr_ok;
}

static jfauderr readpcmdata(jfaud_file *file, jfaud_buffer *data, int samples)
{
	struct vocinfo *voc;
	int samps;
	jfauderr err;
	
	if (file->type != jfaudfile_voc) return jfauderr_unrecognised;

	voc = (struct vocinfo *)file->readerinfo;

	samps = (voc->audiolength - voc->pcmposn) / (voc->channels * voc->bitspersample / 8);
	if (samples < samps) samps = samples;

	if ((err = jfaudbuffer_alloc(data, voc->samplerate, voc->channels, voc->bitspersample, samps)) != jfauderr_ok)
		return err;

	err = readvocsamples(file, data, samps);
	if (err != jfauderr_ok) return err;
	
	if (data->numsamples == 0) jfaudbuffer_free(data);

	return jfauderr_ok;
}

static jfauderr readallpcmdata(jfaud_file *file, jfaud_buffer *data)
{
	struct vocinfo *voc;
	int samps;
	jfauderr err;
	
	if (file->type != jfaudfile_voc) return jfauderr_unrecognised;

	voc = (struct vocinfo *)file->readerinfo;

	samps = (voc->audiolength - voc->pcmposn) / (voc->channels * voc->bitspersample / 8);

	if ((err = jfaudbuffer_alloc(data, voc->samplerate, voc->channels, voc->bitspersample, samps)) != jfauderr_ok)
		return err;

	err = readvocsamples(file, data, samps);
	if (err != jfauderr_ok) return err;
	
	if (data->numsamples == 0) jfaudbuffer_free(data);

	return jfauderr_ok;
}

static jfauderr rewindpcmdata(jfaud_file *file)
{
	struct vocinfo *voc;
	
	if (file->type != jfaudfile_voc) return jfauderr_unrecognised;

	voc = (struct vocinfo *)file->readerinfo;

	voc->pcmposn = 0;
	params->call_seek(file->fh, voc->audiostart, SEEK_SET);
	voc->blocksize = 0;
	voc->blocktype = -1;

	return jfauderr_ok;
}

static jfauderr getpcmlength(jfaud_file *file, unsigned int *samples)
{
	struct vocinfo *voc;
	
	if (file->type != jfaudfile_voc) return jfauderr_unrecognised;

	voc = (struct vocinfo *)file->readerinfo;
	*samples = voc->audiolength / (voc->channels * voc->bitspersample / 8);

	return jfauderr_ok;
}

static jfauderr getplaytime(jfaud_file *file, double *len)
{
	struct vocinfo *voc;
	
	if (file->type != jfaudfile_voc) return jfauderr_unrecognised;

	voc = (struct vocinfo *)file->readerinfo;
	*len = (double)(voc->audiolength / (voc->channels * voc->bitspersample / 8)) / (double)voc->samplerate;

	return jfauderr_ok;
}


jfaud_filereader reader_voc = {
	"Creative VOC",
	jfaudfile_voc,
	jfaudfiledata_pcm,
	
	init,
	uninit,
	testfile,
	cleanup,

	readpcmdata,
	readallpcmdata,
	rewindpcmdata,
	getpcmlength,
	getplaytime
};
