#include "internal.h"

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#ifndef _WIN32
#include <unistd.h>
#endif

#ifndef O_BINARY
#define O_BINARY 0
#endif


static struct jfaud_init_params jfaudparams;
static char initialised = 0;

// these must match the order in the jfaud.h file
jfaud_filereader *filereaders[numfilereaders] = {
	&reader_wave,
	&reader_voc,
	&reader_vorbis,
	&reader_mpeg,
	&reader_kwv,
	
	&reader_midi
};

char filereaderspresent[numfilereaders];


static int stdopen(const char *fn) { return open(fn, O_RDONLY|O_BINARY); }
static int stdclose(int h) { return close(h); }
static int stdread(int h, void *buf, int len) { return read(h, buf, len); }
static int stdseek(int h, int off, int whence) { return lseek(h, off, whence); }
static int stdtell(int h) { return lseek(h,0,SEEK_CUR); }
static void stdlogmsg(const char *str) { fputs(str, stderr); fputc('\n',stderr); }


jfauderr jfaud_init(struct jfaud_init_params *params)
{
	int i;
	jfauderr err;

	if (initialised) return jfauderr_alreadyinited;

	if (params->wavedevice) jfaudparams.wavedevice = strdup(params->wavedevice);
	else jfaudparams.wavedevice = NULL;

	if (params->mididevice) jfaudparams.mididevice = strdup(params->mididevice);
	else jfaudparams.mididevice = NULL;

	if (params->cdadevice) jfaudparams.cdadevice = strdup(params->cdadevice);
	else jfaudparams.cdadevice = NULL;

	jfaudparams.maxvoices = max(4,params->maxvoices);
	
	if (params->samplerate) jfaudparams.samplerate = params->samplerate;
	else jfaudparams.samplerate = 22050;

	if (params->call_open) jfaudparams.call_open = params->call_open;
	else jfaudparams.call_open = stdopen;

	if (params->call_close) jfaudparams.call_close = params->call_close;
	else jfaudparams.call_close = stdclose;

	if (params->call_read) jfaudparams.call_read = params->call_read;
	else jfaudparams.call_read = stdread;

	if (params->call_seek) jfaudparams.call_seek = params->call_seek;
	else jfaudparams.call_seek = stdseek;

	if (params->call_tell) jfaudparams.call_tell = params->call_tell;
	else jfaudparams.call_tell = stdtell;

	if (params->call_logmsg) jfaudparams.call_logmsg = params->call_logmsg;
	else jfaudparams.call_logmsg = stdlogmsg;

	initialised = 1;

	debuglog(jfaudparams.call_logmsg,"main: Initialising file system");
	if ((err = jfaudfile_init()) != jfauderr_ok) {
		jfaud_uninit();
		return err;
	}
	
	debuglog(jfaudparams.call_logmsg,"main: Initialising channel manager");
	if ((err = jfaudchan_init()) != jfauderr_ok) {
		jfaud_uninit();
		return err;
	}

	debuglog(jfaudparams.call_logmsg,"main: Initialising file readers");
	jfaudparams.call_logmsg(jsprintf("Supported file formats:"));
	for (i=0; i<numfilereaders; i++) {
		filereaderspresent[i] = 1;
		if ((err = filereaders[i]->init()) != jfauderr_ok) {
			if (err == jfauderr_notinited) {
				jfaudparams.call_logmsg(jsprintf("  %-15s (no -- failed to initialise)",filereaders[i]->name));
				filereaderspresent[i] = 0;	// disable the reader
			} else if (err == jfauderr_unrecognised) {
				jfaudparams.call_logmsg(jsprintf("  %-15s (disabled)",filereaders[i]->name));
				filereaderspresent[i] = 0;	// disable the reader
			} else {
				jfaud_uninit();
				return err;
			}
		} else {
			jfaudparams.call_logmsg(jsprintf("  %-15s (yes)",filereaders[i]->name));
		}
	}

	return jfauderr_ok;
}


jfauderr jfaud_uninit(void)
{
	int i;
	
	if (!initialised) return jfauderr_ok;

	// stop the channels and file services before unitialising the readers
	jfaudchan_uninit();
	jfaudfile_uninit();

	for (i=0; i<numfilereaders; i++) {
		if (filereaderspresent[i]) filereaders[i]->uninit();
	}
	
	if (jfaudparams.wavedevice) free(jfaudparams.wavedevice);
	if (jfaudparams.mididevice) free(jfaudparams.mididevice);
	if (jfaudparams.cdadevice) free(jfaudparams.cdadevice);

	jfaudparams.wavedevice = NULL;
	jfaudparams.mididevice = NULL;
	jfaudparams.cdadevice = NULL;

	initialised = 0;

	return jfauderr_ok;
}


jfauderr jfaud_isinited(void)
{
	if (initialised) return jfauderr_ok;
	return jfauderr_notinited;
}


jfauderr jfaud_update(void)
{
	jfauderr err;

	if ((err = jfaudchan_update()) != jfauderr_ok) return err;

	return jfauderr_ok;
}


struct jfaud_init_params * jfaud_getparams(void)
{
	if (initialised) return &jfaudparams;
	return NULL;
}


#include <stdarg.h>
#ifdef _MSC_VER
#define vsnprintf _vsnprintf
#endif
static char sprintfbuf[1024];
char *jsprintf(const char *fmt, ...)
{
	va_list va;
	va_start(va,fmt);
	vsnprintf(sprintfbuf,1024,fmt,va);
	va_end(va);
	return sprintfbuf;
}


const char *jfauderr_getstring(jfauderr err)
{
	switch (err) {
		case jfauderr_ok: return "Ok.";

		case jfauderr_genericerror: return "Generic error.";
		case jfauderr_outofmemory: return "Out of memory.";

		case jfauderr_notinited: return "Not initialised.";
		case jfauderr_alreadyinited: return "Already initialised.";

		case jfauderr_openerror: return "Open error.";

		case jfauderr_unrecognised: return "Unrecognised format.";
		case jfauderr_wrongtype: return "Wrong file type.";
		case jfauderr_nowave: return "No waveform output.";
		case jfauderr_nomidi: return "No MIDI output.";
		case jfauderr_nocda: return "No CD-Audio output.";

		case jfauderr_notplaying: return "Not playing.";
		case jfauderr_notloaded: return "No file loaded.";
		case jfauderr_notprepared: return "Buffer not prepared.";

		case jfauderr_badhandle: return "Bad handle.";

		default: return "Unknown error.";
	}
}

