#ifndef _WIN32
#error This is only for Win32 targets
#endif

#if defined _WIN32 && defined PACKED
#undef PACKED
#endif
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <mmsystem.h>
#include <stdlib.h>
#include <ctype.h>

#include "internal.h"

#ifndef MOD_WAVETABLE
#define MOD_WAVETABLE 6
#define MOD_SWSYNTH 7
#endif

#define STREAMBUFLEN	4096 // * 4 bytes = 16K max per 1/4 note
#define NUMMIDIBUFS 2

typedef struct {
	HMIDISTRM stream;
	MIDIHDR header[NUMMIDIBUFS];
	DWORD bufdata[NUMMIDIBUFS][STREAMBUFLEN], buflen[NUMMIDIBUFS];
	int curbuf, running, ppqn, filler;
} MidiOut;

static mididevtype midicardtype;
static int mididevicenum = -1;
static MidiOut *curmidiout = NULL;

static void releasemidiout(MidiOut *out)
{
	midiStreamStop(out->stream);
	midiStreamClose(out->stream);
	free(out);
}

jfauderr midiout_init(void)
{
	int numdevs, i, j, matchdev = -1, matchchars = -1;
	MIDIOUTCAPS caps, matchcaps;
	char *devtype;

	// figure out which midi device to select, and what type of card it resembles
	if (!conf.mididev) return jfauderr_noerror;

	numdevs = midiOutGetNumDevs();
	if (!numdevs) {
		logmsg("No MIDI devices available.");
		return jfauderr_notinited;
	}

	logmsg("Available MIDI devices:");
	if (!conf.mididev[0]) matchdev=0;
	for (i=0; i<numdevs; i++) {
		if (midiOutGetDevCaps(i,&caps,sizeof(MIDIOUTCAPS)) != MMSYSERR_NOERROR) continue;

		for (j=0; caps.szPname[j] && conf.mididev[j] &&
			toupper(caps.szPname[j]) == toupper(conf.mididev[j]);
			j++) ;
		if (j > matchchars)
			matchchars = j, matchdev = i, memcpy(&matchcaps,&caps, sizeof(MIDIOUTCAPS));

		logmsg(jsprintf("  - %s",caps.szPname));
	}
	switch (matchcaps.wTechnology) {
		case MOD_MIDIPORT:
			devtype = "a MIDI hardware port.";
			midicardtype = mididevtype_genmidi;
			break;
		case MOD_SYNTH:
			devtype = "a generic synthesiser.";
			midicardtype = mididevtype_genmidi;
			break;
		case MOD_SQSYNTH:
			devtype = "a square-wave synthesiser.";
			midicardtype = mididevtype_fmopl2;
			break;
		case MOD_FMSYNTH:
			devtype = "an FM synthesiser.";
			midicardtype = mididevtype_fmopl2;
			break;
		case MOD_WAVETABLE:
			devtype = "a wavetable synthesiser.";
			midicardtype = mididevtype_genmidi;
			break;
		case MOD_SWSYNTH:
			devtype = "a software synthesiser.";
			midicardtype = mididevtype_genmidi;
			break;
		default:
			devtype = "some unusual type of synthesiser.";
			midicardtype = mididevtype_genmidi;
			break;
	}
	logmsg(jsprintf("Selected device \"%s\" is %s",matchcaps.szPname,devtype));
	mididevicenum = matchdev;

	curmidiout = NULL;

	return jfauderr_noerror;
}

jfauderr midiout_uninit(void)
{
	if (curmidiout) {
		releasemidiout(curmidiout);
		curmidiout = NULL;
	}

	mididevicenum = -1;

	return jfauderr_noerror;
}

static void CALLBACK midiCallback(HMIDIOUT handle, UINT uMsg, DWORD dwInstance, DWORD dwParam1, DWORD dwParam2 UNUSED)
{
        LPMIDIHDR hdr = (LPMIDIHDR)dwParam1;
	MidiOut *out = (MidiOut *)dwInstance;

        if (uMsg != MOM_DONE) return;

        midiOutUnprepareHeader(handle,hdr,sizeof(MIDIHDR));
	out->buflen[hdr->dwUser] = 0;
}

jfauderr midiout_allocdevice(int *handle)
{
	MidiOut *out;
	MMRESULT ret;
	
	if (mididevicenum < 0) return jfauderr_notinited;
	if (curmidiout) return jfauderr_deverror;
	if (!handle) return jfauderr_genericerror;

	out = (MidiOut*)calloc(1,sizeof(MidiOut));
	if (!out) return jfauderr_outofmemory;

	out->curbuf = -1;

	ret = midiStreamOpen(&out->stream, &mididevicenum, 1, (DWORD)midiCallback, (DWORD)out, CALLBACK_FUNCTION);
	if (ret != MMSYSERR_NOERROR) {
		free(out);
		return jfauderr_deverror;
	}

	curmidiout = out;
	*handle = (int)curmidiout;
	return jfauderr_noerror;
}

jfauderr midiout_releasedevice(int handle)
{
	if (mididevicenum < 0) return jfauderr_notinited;
	if ((int)curmidiout != handle || !curmidiout) return jfauderr_badhandle;

	releasemidiout(curmidiout);
	curmidiout = NULL;

	return jfauderr_noerror;
}

jfauderr midiout_settimebase(int handle, int ppqn)
{
	MIDIPROPTIMEDIV tdiv;
	MMRESULT ret;

	if (mididevicenum < 0) return jfauderr_notinited;
	if ((int)curmidiout != handle || !curmidiout) return jfauderr_badhandle;

	tdiv.cbStruct = sizeof(MIDIPROPTIMEDIV);
	tdiv.dwTimeDiv = ppqn;
	ret = midiStreamProperty(curmidiout->stream, (LPBYTE)&tdiv, MIDIPROP_SET|MIDIPROP_TIMEDIV);
	if (ret != MMSYSERR_NOERROR) return jfauderr_deverror;

	curmidiout->ppqn = ppqn;

	return jfauderr_noerror;
}

jfauderr midiout_getdevicetype(int handle, mididevtype *type)
{
	if (mididevicenum < 0) return jfauderr_notinited;
	if ((int)curmidiout != handle || !curmidiout) return jfauderr_badhandle;

	*type = midicardtype;
	return jfauderr_noerror;
}

static int findfreebuf(MidiOut *out)
{
	int i;

	for (i=NUMMIDIBUFS-1; i>=0; i--)
		if (out->buflen[i] == 0)
			return i;
	return -1;
}

jfauderr midiout_needsdata(int handle)
{
	if (mididevicenum < 0) return jfauderr_notinited;
	if ((int)curmidiout != handle || !curmidiout) return jfauderr_badhandle;

	return (findfreebuf(curmidiout) < 0) ? jfauderr_genericerror : jfauderr_noerror;
}

jfauderr midiout_bufferevent(int handle, unsigned delta, unsigned char cmd, unsigned len, unsigned char const *data)
{
	DWORD *edata, elen;
	
	if (mididevicenum < 0) return jfauderr_notinited;
	if ((int)curmidiout != handle || !curmidiout) return jfauderr_badhandle;
	if (len>0 && !data) return jfauderr_genericerror;

	if (curmidiout->curbuf < 0) {
		curmidiout->curbuf = findfreebuf(curmidiout);
		if (curmidiout->curbuf < 0) return jfauderr_genericerror;	// no buffers ready
	} else {
		int evtlen = 2;

		if (cmd == 0 || len == 0) evtlen++;	// NOP
		else if (cmd == 0xff && data[0] == 0x51) evtlen++;	// tempo
		else if (len < 3) evtlen++;		// short
		else evtlen += 1 + ((len+1)+3)/4;	// long
				
		if (curmidiout->buflen[ curmidiout->curbuf ] + evtlen > STREAMBUFLEN)
			return jfauderr_buffull;	// buffer is full
	}

	edata = curmidiout->bufdata[ curmidiout->curbuf ];
	elen  = curmidiout->buflen[ curmidiout->curbuf ];

	edata[ elen++ ] = delta;
	edata[ elen++ ] = 0;
	if ( cmd == 0 || len == 0 )
		edata[ elen++ ] = (MEVT_NOP<<24);
	else if (cmd == 0xff && data[0] == 0x51)
		edata[ elen++ ] = (MEVT_TEMPO<<24)|((DWORD)data[2]<<16)|((DWORD)data[3]<<8)|((DWORD)data[4]);
	else if (len < 3) {
		edata[ elen ] = MEVT_F_SHORT | cmd;
		switch (len) {
			case 2: edata[ elen ] |= (DWORD)data[1] << 16;
			case 1: edata[ elen ] |= (DWORD)data[0] << 8;
			default: break;
		}
		elen++;
	} else {
		unsigned j,k;
		edata[ elen++ ] = MEVT_F_LONG | (1+len);
		edata[ elen ] = cmd;
		for(j=1,k=0;k<len;k++) {
			edata[ elen ] |= ((DWORD)data[k] << (j*8));
			if (++j==4) {
				j=0;
				edata[ elen++ ] = 0;
			}
		}
		if (j>0) elen++;	// ended on a partial?
	}

	curmidiout->buflen[ curmidiout->curbuf ] = elen;

	return jfauderr_noerror;
}

jfauderr midiout_finish(int handle)
{
	if (mididevicenum < 0) return jfauderr_notinited;
	if ((int)curmidiout != handle || !curmidiout) return jfauderr_badhandle;

	if (curmidiout->curbuf >= 0) {
		int num = curmidiout->curbuf;
		MMRESULT ret;

		curmidiout->header[num].lpData = (LPSTR)&curmidiout->bufdata[num];
		curmidiout->header[num].dwBufferLength = 
		curmidiout->header[num].dwBytesRecorded = curmidiout->buflen[num]*4;
		curmidiout->header[num].dwUser = num;
		curmidiout->header[num].dwFlags = 0;
		
		ret = midiOutPrepareHeader((HMIDIOUT)curmidiout->stream, &curmidiout->header[num], sizeof(MIDIHDR));
		if (ret != MMSYSERR_NOERROR) return jfauderr_genericerror;
		ret = midiStreamOut(curmidiout->stream, &curmidiout->header[num], sizeof(MIDIHDR));
		if (ret != MMSYSERR_NOERROR) return jfauderr_genericerror;
		
		curmidiout->curbuf = -1;
	}

	if (!curmidiout->running) {
		MMRESULT ret;
		ret = midiStreamRestart(curmidiout->stream);
	        if (ret != MMSYSERR_NOERROR) return jfauderr_deverror;

		curmidiout->running = 1;
	}

	return jfauderr_noerror;
}

