#include <stdlib.h>
#include <string.h>

#include "internal.h"

int sysf_open(const char *fn, const char *subfn, JFAudFH *h, JFAudRawFormat **raw)
{
	if (!h) return -1;
	if (h->fh < 0) return -1;

	return conf.cb_open(fn,subfn,h,raw);
}

int sysf_close(JFAudFH *h)
{
	if (!h) return -1;
	if (h->fh < 0) return -1;

	return conf.cb_close(h);
}

int sysf_read(JFAudFH *h, void *buf, int len)
{
	if (!h) return -1;
	if (h->fh < 0) return -1;

	return conf.cb_read(h,buf,len);
}

int sysf_seek(JFAudFH *h, int off, int whence)
{
	if (!h) return -1;
	if (h->fh < 0) return -1;

	return conf.cb_seek(h,off,whence);
}

int sysf_tell(JFAudFH *h)
{
	if (!h) return -1;
	if (h->fh < 0) return -1;

	return conf.cb_tell(h);
}

int sysf_filelen(JFAudFH *h)
{
	if (!h) return -1;
	if (h->fh < 0) return -1;

	return conf.cb_filelen(h);
}


#include "filereaders.h"
#define numfilereaders (int)(sizeof(filereaders)/sizeof(filereaders[0]))

jfauderr filereaders_init(void)
{
	jfauderr err;
	int i;
	
	logmsg("Initialising file readers...");
	for (i=0; i<numfilereaders; i++) {
		filereaders[i].inited = 0;
		err = filereaders[i].reader->init();
		if (err == jfauderr_noerror) {
			logmsg(jsprintf("  %-15s", filereaders[i].reader->name));
			filereaders[i].inited = 1;
		} else if (err == jfauderr_disabled) {
			logmsg(jsprintf("  %-15s (disabled)", filereaders[i].reader->name));
		} else if (err == jfauderr_initfailed) {
			logmsg(jsprintf("  %-15s (failed to initialise)", filereaders[i].reader->name));
		} else {
			return err;
		}
	}

	return jfauderr_noerror;
}

jfauderr filereaders_uninit(void)
{
	jfauderr err;
	int i;
	
	for (i=0; i<numfilereaders; i++) {
		if (!filereaders[i].inited) continue;
		err = filereaders[i].reader->uninit();
	}

	return jfauderr_noerror;
}

jfauderr filereaders_identify(File *file)
{
	jfauderr err;
	int i;

	for (i=0; i<numfilereaders; i++) {
		if (!filereaders[i].inited) continue;
		err = filereaders[i].reader->test(file);
		if (err == jfauderr_noerror) {
			file->readeridx = i;
			debuglog(DEBUGLVL_GENERAL,jsprintf("Identified as a %s", filereaders[i].reader->name));
			return jfauderr_noerror;	// identified and ready to roll
		} else if (err == jfauderr_notrecognised || err == jfauderr_notinited) {
			continue;
		} else {
			return err;
		}
	}

	return jfauderr_notrecognised;
}

jfauderr filereaders_unload(File *file, char *noclose)
{
	if (!file) return jfauderr_genericerror;
	if (!file->sf) return jfauderr_genericerror;

	if (file->readeridx < 0) return wf_raw.unload(file, noclose);
	
	return filereaders[ file->readeridx ].reader->unload(file, noclose);
}

jfauderr filereaders_wave_readsamples(File *fil, int n, PcmBuffer *buf, char loop)
{
	if (!fil) return jfauderr_genericerror;
	if (!fil->sf) return jfauderr_genericerror;
	if (fil->sf->type != type_wave) return jfauderr_notcompatible;

	if (fil->readeridx < 0) return wf_raw.readsamples(fil,n,buf,loop);

	return ((WaveformFileReader*)filereaders[ fil->readeridx ].reader)->readsamples(fil,n,buf,loop);
}

jfauderr filereaders_wave_readallsamples(File *fil, PcmBuffer *buf)
{
	if (!fil) return jfauderr_genericerror;
	if (!fil->sf) return jfauderr_genericerror;
	if (fil->sf->type != type_wave) return jfauderr_notcompatible;

	if (fil->readeridx < 0) return wf_raw.readallsamples(fil,buf);

	return ((WaveformFileReader*)filereaders[ fil->readeridx ].reader)->readallsamples(fil,buf);
}

jfauderr filereaders_wave_getplaytime(File *fil, float *len)
{
	if (!fil) return jfauderr_genericerror;
	if (!fil->sf) return jfauderr_genericerror;
	if (fil->sf->type != type_wave) return jfauderr_notcompatible;

	if (fil->readeridx < 0) return wf_raw.getplaytime(fil,len);

	return ((WaveformFileReader*)filereaders[ fil->readeridx ].reader)->getplaytime(fil,len);
}

jfauderr filereaders_midi_getdata(File *fil, MidiBuffer *buf)
{
	if (!fil) return jfauderr_genericerror;
	if (!fil->sf) return jfauderr_genericerror;
	if (fil->sf->type != type_midi) return jfauderr_notcompatible;

	return ((MidiFileReader*)filereaders[ fil->readeridx ].reader)->getdata(fil,buf);
}

jfauderr soundfile_open(File **fh, const char *fn, const char *subfn)
{
	jfauderr err;
	File *nfh;
	JFAudRawFormat *raw = NULL;

	if (!fh) return jfauderr_badhandle;
	if (!fn) return jfauderr_genericerror;

	nfh = (File *)calloc(1,sizeof(File));
	if (!nfh) return jfauderr_outofmemory;

	if (sysf_open(fn,subfn,&nfh->fh,&raw) < 0) {
		free(nfh);
		return jfauderr_openerror;
	}

	nfh->filename = (subfn ? strdup(subfn) : (fn ? strdup(fn) : NULL));

	if (raw) {
		nfh->readeridx = -1;		// we don't go through the usual channels
		nfh->sf = (SoundFile*)raw;	// because there's no real way to pass
						// something into wf_raw->test(), we hack around it
		err = wf_raw.test(nfh);
		if (err != jfauderr_noerror) {
			sysf_close(&nfh->fh);
			free(nfh);
		} else {
			nfh->ident = hashfn(fn,subfn);
			*fh = nfh;
		}
		return err;
	}
	
	err = filereaders_identify(nfh);
	if (err != jfauderr_noerror) {
		sysf_close(&nfh->fh);
		free(nfh);
	} else {
		nfh->ident = hashfn(fn,subfn);
		*fh = nfh;
	}
	return err;
}

jfauderr soundfile_close(File *fh)
{
	char noclose = 0;
	jfauderr err = jfauderr_noerror;
	if (!fh) return jfauderr_badhandle;
	if (fh->sf) {
		err = filereaders_unload(fh, &noclose);
		free(fh->sf);
	}
	if (!noclose) sysf_close(&fh->fh);
	if (fh->filename) free(fh->filename);
	free(fh);
	return err;
}

